/*
+-------------------------------- DISCLAIMER ---------------------------------+
|                                                                             |
| This application program is provided to you free of charge as an example.   |
| Despite the considerable efforts of Euresys personnel to create a usable    |
| example, you should not assume that this program is error-free or suitable  |
| for any purpose whatsoever.                                                 |
|                                                                             |
| EURESYS does not give any representation, warranty or undertaking that this |
| program is free of any defect or error or suitable for any purpose. EURESYS |
| shall not be liable, in contract, in torts or otherwise, for any damages,   |
| loss, costs, expenses or other claims for compensation, including those     |
| asserted by third parties, arising out of or in connection with the use of  |
| this program.                                                               |
|                                                                             |
+-----------------------------------------------------------------------------+
*/

// Helper for Audio Video Codec
// 2.0.0.15

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdexcept>
#include <memory>
#include <poll.h>
#include <sys/types.h>
#include <sys/mman.h>

extern "C" {
#include <avcodec.h>
};



namespace vfd_euresys
{


	// class managing the H264 decompression
	class AvcodecH264Decoder
	{
	public:
		AvcodecH264Decoder() : codec(NULL), codecContext(NULL), picture(NULL)
		{
			avcodec_register_all();
			Initialize();
		}

		~AvcodecH264Decoder()
		{
			av_free(picture);
			if (codecContext)
				avcodec_close(codecContext);
			av_free(codecContext);
		}

		/** Returns a pointer to the frame that was decoded if any.
		The pointer is valid until the next call to this function.
		If more than one frame was decoded, only the last one is given.
		Returns NULL if no frame was decoded and throws on error.
		**/
		AVFrame *Decode(char *data, size_t size)
		{
			int gotPicture, len;
			AVFrame *frame = NULL;

			do {
				AVPacket avpkt;
				av_init_packet(&avpkt);
				avpkt.data = (uint8_t *)data;
				avpkt.size = size;
				avpkt.flags = AV_PKT_FLAG_KEY;

				len = avcodec_decode_video2(codecContext, picture, &gotPicture, &avpkt);

				if (len < 0) {
					throw std::runtime_error("Error while decoding frame");
				}
				if (gotPicture) {
					frame = picture;
				}
				size -= len;
				data += len;
			} while (size > 0);
			return frame;
		}
	void Flush()
	{
		avcodec_flush_buffers(codecContext);
	}

	private:
		void Initialize()
		{
			codec = avcodec_find_decoder(CODEC_ID_H264);
			if (!codec) {
				throw std::runtime_error("codec not found");
			}
			codecContext = avcodec_alloc_context3(codec);
			if (!codecContext) {
				throw std::runtime_error("could not create codec context");
			}
			if (codec->capabilities & CODEC_CAP_TRUNCATED) {
				codecContext->flags |= CODEC_FLAG_TRUNCATED;
			}
			if (avcodec_open2(codecContext, codec, NULL) < 0) {
				av_free(codecContext);
				throw std::runtime_error("could not open codec");
			}
			codecContext->flags2 |= CODEC_FLAG2_CHUNKS;
			picture = avcodec_alloc_frame();
			if (!picture) {
				avcodec_close(codecContext);
				av_free(codecContext);
				throw std::runtime_error("could not allocate frame");
			}
		}

		AVCodec *codec;
		AVCodecContext *codecContext;
		AVFrame *picture;
	};

};

