/*
+-------------------------------- DISCLAIMER ---------------------------------+
|                                                                             |
| This application program is provided to you free of charge as an example.   |
| Despite the considerable efforts of Euresys personnel to create a usable    |
| example, you should not assume that this program is error-free or suitable  |
| for any purpose whatsoever.                                                 |
|                                                                             |
| EURESYS does not give any representation, warranty or undertaking that this |
| program is free of any defect or error or suitable for any purpose. EURESYS |
| shall not be liable, in contract, in torts or otherwise, for any damages,   |
| loss, costs, expenses or other claims for compensation, including those     |
| asserted by third parties, arising out of or in connection with the use of  |
| this program.                                                               |
|                                                                             |
+-----------------------------------------------------------------------------+
*/

// 2.0.0.15

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <cstdlib>
#include <stdio.h>
#include <string.h>

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/xmlwriter.h>

#include "tiffio.h"


using namespace std;

enum Itype
{
	RGB24,
	RGB32,
	RGB15,
	RGB16,
	Y8,
	YUV422,
	YUV422PL,
	YUV420PL

};

int minmax ( int value, int min, int max )
{
	int val;
	val = value > max ? max : value;
	val = value < min ? min : val;
	return val;
}

// function to browse to the next xml node
xmlNodePtr GetNextNode ( xmlNodePtr node )
{
	if (NULL == node)
	{
	    printf("\nXml struct has changed, sample is not able to deal with installed VFD version\n");
		exit(-1);
	}

	xmlNodePtr cur;
	cur = node->next;
	if ( cur->type == XML_TEXT_NODE )
		cur = cur->next;

	if (NULL == cur)
	{
	    printf("\nXml struct has changed, sample is not able to deal with installed VFD version\n");
		exit(-1);
	}

	return cur;
}
// function to browse to the next child node
xmlNodePtr GetNextChild ( xmlNodePtr node )
{
	xmlNodePtr cur;
	if (NULL == node)
	{
	    printf("\nXml struct has changed, sample is not able to deal with installed VFD version\n");
		exit(-1);
	}

	cur = node->xmlChildrenNode;;
	if ( cur->type == XML_TEXT_NODE )
		cur = cur->next;

	return cur;
}

// read data from config.xml file

int ReadConfigData ( char * configfile, int * sizeX, int * sizeY, Itype * color)
{
	xmlChar * ctmp;
	int count;
	// xml document & node pointer
	xmlDocPtr doc;
	xmlNodePtr cur;

	// read the file
	doc = xmlParseFile ( configfile );

	if ( doc == NULL )
	{
		fprintf ( stderr, "Could not open %s file\n",configfile );
		exit ( -1 );

	}
	// get file type
	cur = xmlDocGetRootElement ( doc );
	if ( cur == NULL )
	{
		xmlFreeDoc ( doc );
		fprintf ( stderr, "Could not find root element\n",configfile );
		exit ( -1 );
	}

	if ( xmlStrcmp ( cur->name, BAD_CAST "VisualFormatter" ) )
	{
		xmlFreeDoc ( doc );
		fprintf ( stderr, "Could not find Visual Formatter root element\n",configfile );
		exit ( -1 );
	}


	// verbose/concise option
	cur = GetNextChild ( cur );
	if ( xmlStrcmp ( cur->name, BAD_CAST "Verbose" ) )
	{
		xmlFreeDoc ( doc );
		fprintf ( stderr, "Config file is not in verbose mode\n",configfile );
		exit ( -1 );
	}
	// parse elements
	// active/idle
	cur = GetNextNode ( cur );

	// decimation
	cur = GetNextNode ( cur );
	// format type
	cur = GetNextNode ( cur );
	if ( ! xmlStrcmp ( cur->name, BAD_CAST "FormatRGB32" ) ) * color = RGB32;
	if ( ! xmlStrcmp ( cur->name, BAD_CAST "FormatRGB24" ) ) * color = RGB24;
	if ( ! xmlStrcmp ( cur->name, BAD_CAST "FormatRGB16" ) ) * color = RGB16;
	if ( ! xmlStrcmp ( cur->name, BAD_CAST "FormatRGB15" ) ) * color = RGB15;
	if ( ! xmlStrcmp ( cur->name, BAD_CAST "FormatY8" ) ) * color = Y8;
	if ( ! xmlStrcmp ( cur->name, BAD_CAST "FormatYUV422" ) ) * color = YUV422;
	if ( ! xmlStrcmp ( cur->name, BAD_CAST "FormatYUV422PL" ) ) * color = YUV422PL;
	if ( ! xmlStrcmp ( cur->name, BAD_CAST "FormatYUV420PL" ) ) * color = YUV420PL;

	// scaling
	cur = GetNextNode ( cur );

	// sizeX
	cur = GetNextNode ( cur );
	ctmp = xmlNodeListGetString ( doc, cur->xmlChildrenNode, 1 );
	sscanf ( ( const char* ) ctmp,"%d",sizeX );

	// sizeY
	cur = GetNextNode ( cur );
	ctmp = xmlNodeListGetString ( doc, cur->xmlChildrenNode, 1 );
	sscanf ( ( const char* ) ctmp,"%d",sizeY );

	// finish with this metadata
	xmlFreeDoc ( doc );

	return count;

}

// convert RGB15 to RGB24 color format

void RGB15toRGB24 ( char * source, char * dest, int SizeX, int SizeY )
{
	int i,j,pscr, pdes;
	for ( j=0; j<SizeY; j++ )
	{
		for ( i=0; i<SizeX; i++ )
		{
			pscr = ( i+j*SizeX ) *2;
			pdes = ( i+j*SizeX ) *3;
			dest[pdes] = ( source[pscr+1] & 0x7C ) *2;
			dest[pdes+1] = ( source[pscr+1] & 0x03 ) *64 + ( source[pscr] & 0xE0 ) /4;
			dest[pdes+2] = ( source[pscr] & 0x1F ) *8;
		}

	}

}

// convert RGB24 to RGB24 color format (swap red and blue)

void RGB24toRGB24 ( char * source, char * dest, int SizeX, int SizeY )
{
	int i,j,pscr, pdes;
	for ( j=0; j<SizeY; j++ )
	{
		for ( i=0; i<SizeX; i++ )
		{
			pscr = ( i+j*SizeX ) *3;
			pdes = ( i+j*SizeX ) *3;
			dest[pdes] = source[pscr+2];
			dest[pdes+1] = source[pscr+1];
			dest[pdes+2] = source[pscr];
		}

	}

}

// convert RGB16 to RGB24 color format

void RGB16toRGB24 ( char * source, char * dest, int SizeX, int SizeY )
{
	int i,j,pscr, pdes;
	for ( j=0; j<SizeY; j++ )
	{
		for ( i=0; i<SizeX; i++ )
		{
			pscr = ( i+j*SizeX ) *2;
			pdes = ( i+j*SizeX ) *3;
			dest[pdes] = ( source[pscr+1] & 0xF8 );
			dest[pdes+1] = ( source[pscr+1] & 0x07 ) * 32 + ( source[pscr] & 0xE0 ) /8;
			dest[pdes+2] = ( source[pscr] & 0x1F ) *8;
		}

	}

}

// convert RGB32 to RGB24 color format

void RGB32toRGB24 ( char * source, char * dest, int SizeX, int SizeY )
{
	int i,j,pscr, pdes;
	for ( j=0; j<SizeY; j++ )
	{
		for ( i=0; i<SizeX; i++ )
		{
			pscr = ( i+j*SizeX ) *4;
			pdes = ( i+j*SizeX ) *3;
			dest[pdes] = source[pscr+2];
			dest[pdes+1] = source[pscr+1];
			dest[pdes+2] = source[pscr];
		}

	}
}

// convert grey scale RGB24 color format

void Y8toRGB24 ( char * source, char * dest, int SizeX, int SizeY )
{
	int i,j,pscr, pdes;
	for ( j=0; j<SizeY; j++ )
	{
		for ( i=0; i<SizeX; i++ )
		{
			pscr = ( i+j*SizeX );
			pdes = ( i+j*SizeX ) *3;
			dest[pdes] = source[pscr];
			dest[pdes+1] = source[pscr];
			dest[pdes+2] = source[pscr];
		}

	}
}

// convert YUV 4:2:2 to RGB24 color format

void YUV422toRGB24 ( char * source, char * dest, int SizeX, int SizeY )
{
	int i,j,pscr, pdes;
	float y1,y2,u,v;
	int a1, b1, c1, a2, b2, c2;

	for ( j=0; j<SizeY; j++ )
	{
		for ( i=0; i<SizeX; i++ )
		{
			pscr = ( i+j*SizeX ) *2;
			pdes = ( i+j*SizeX ) *3;
			// take into account YUV offset
			y1 = ( unsigned char ) source[pscr]-16;
			y2 = ( unsigned char ) source[pscr+2]-16 ;
			u = ( unsigned char ) source[pscr+1] -128;
			v = ( unsigned char ) source[pscr+3] -128;
			// YUV to RGB matrix
			// handle 2 pixels at a time
			a1 = 1.164F*y1 + 1.596F*v;
			b1 = 1.164F*y1 - 0.813F*v- 0.392F*u;
			c1 = 1.164F*y1 + 2.017F*u;
			a2 = 1.164F*y2 + 1.596F*v;
			b2 = 1.164F*y2 - 0.813F*v- 0.392F*u;
			c2 = 1.164F*y2 + 2.017F*u;

			// make sur not overrun happened
			a1 = minmax ( a1, 0,255 );
			b1 = minmax ( b1, 0,255 );
			c1 = minmax ( c1, 0,255 );
			a2 = minmax ( a2, 0,255 );
			b2 = minmax ( b2, 0,255 );
			c2 = minmax ( c2, 0,255 );

			dest[pdes] =  a1;
			dest[pdes+1] = b1;
			dest[pdes+2] = c1;
			dest[pdes+3] = a2;
			dest[pdes+4] = b2;
			dest[pdes+5] = c2;

			i++;
		}

	}
}

// convert YUV 4:2:2 planar to RGB24 color format

void YUV422PLtoRGB24 ( char * source, char * dest, int SizeX, int SizeY )
{
	int i,j,pscr, pdes, pcb, pcr;
	int a1,b1,c1;
	float y, u,v;
	for ( j=0; j<SizeY; j++ )
	{
		for ( i=0; i<SizeX; i++ )
		{
			pscr = ( i+j*SizeX );
			pdes = ( i+j*SizeX ) *3;
			pcb = ( i+j*SizeX ) /2 +SizeX*SizeY;
			pcr = ( i+j*SizeX ) /2 +SizeX*SizeY*1.5;

			// take into account YUV offset
			y = ( unsigned char ) source[pscr]-16;
			u = ( unsigned char ) source[pcb] -128;
			v = ( unsigned char ) source[pcr] -128;
			// YUV to RGB matrix
			a1 = 1.164F*y + 1.596F*v;
			b1 = 1.164F*y - 0.813F*v- 0.392F*u;
			c1 = 1.164F*y + 2.017F*u;
			// make sur not overrun happened
			a1 = minmax ( a1,0,255 );
			b1 = minmax ( b1,0,255 );
			c1 = minmax ( c1,0,255 );

			dest[pdes] = a1;
			dest[pdes+1] = b1;
			dest[pdes+2] = c1;
		}

	}
}

// convert YUV 4:2:0 planar to RGB24 color format

void YUV420PLtoRGB24 ( char * source, char * dest, int SizeX, int SizeY )
{
	int i,j,pscr, pdes, pcb, pcr;
	int a1,b1,c1;
	float y, u,v;
	for ( j=0; j<SizeY; j++ )
	{
		for ( i=0; i<SizeX; i++ )
		{
			pscr = ( i+j*SizeX );
			pdes = ( i+j*SizeX ) *3;
			pcb = ( i+ ( ( int ) ( j/2 ) ) *SizeX ) /2 +SizeX*SizeY;
			pcr = ( i+ ( ( int ) ( j/2 ) ) *SizeX ) /2 +SizeX*SizeY*1.25;
			// take into account YUV offset
			y = ( unsigned char ) source[pscr]-16;
			u = ( unsigned char ) source[pcb] -128;
			v = ( unsigned char ) source[pcr] -128;
			// YUV to RGB matrix
			a1 = 1.164F*y + 1.596F*v;
			b1 = 1.164F*y - 0.813F*v- 0.392F*u;
			c1 = 1.164F*y + 2.017F*u;
			// make sur not overrun happened
			a1 = minmax ( a1,0,255 );
			b1 = minmax ( b1,0,255 );
			c1 = minmax ( c1,0,255 );

			dest[pdes] = a1;
			dest[pdes+1] = b1;
			dest[pdes+2] = c1;
		}

	}
}

// display the expected arguments

void DisplayUsage()
{
    printf("\nUsage\n");
    printf("=====\n\n");
    printf("raw2tiff mount_point source_id picture_id tiffimage\n\n");
    printf("where\n");
    printf("mount_point is the path to the board's virtual file system\n");
    printf("e.g.: /mnt/u4_0 \n\n");
    printf("source_id is the ID of the target source video (integer)\n");
    printf("e.g.: 0\n\n");
    printf("picture_id is the name of the target source image (integer)\n");
    printf("e.g.: 0\n\n");
    printf("tiffimage is name and path of the tiff that will be saved\n");
    printf("e.g. : /home/myaccount/myimage.tif\n");

}

int main ( int argc, char *argv[] )
{
	int SizeX;       // image size
	int SizeY;
	float Bpp;       // byte per pixel
	char * imgscr;   // Image holding data from the board
	char * imgdes;   // Image holding the data to write to the tiff
	Itype imgtype;   // type of color format to use
        
        char cfgName[256];
        char rawName[256];
        char * tifName;

	TIFF * tif;     // handle to the tiff file
	FILE * raw;     // handle to the VFS file

        // check arguments count
        if (argc < 5)
        {
            DisplayUsage();
            return -1;
        }

	// build path to file system
        sprintf(cfgName,"%s/visual/source%s/formatter/config.xml",argv[1],argv[2]);
        sprintf(rawName,"%s/visual/source%s/formatter/data/%s",argv[1],argv[2],argv[3]);
        tifName = argv[4];
                

	// get imgscr info
	ReadConfigData ( cfgName, &SizeX, &SizeY, & imgtype);
	
	// compute amount of bytes per pixel from the color format information
	switch (imgtype)
	{
	case RGB32 : Bpp = 4;
	   break;
	case RGB24 : Bpp = 3;
	   break;
	case RGB15 : Bpp = 2;
	   break;
	case RGB16 : Bpp = 2;
	   break;
	case Y8 : Bpp = 1;
	   break;
	case YUV422 : Bpp = 2;
	   break;
	case YUV422PL : Bpp = 2;
	   break;
	case YUV420PL : Bpp = 1.5;
	   break;
	
	
	}
	// alloc scr and des
	imgscr = ( char * ) malloc ( sizeof ( char ) * SizeX * SizeY * Bpp );
	if ( !imgscr )
	{
		fprintf ( stderr, "Could not allocate enough memory for source image\n" );
		exit ( -1 );

	}
	imgdes = ( char * ) malloc ( sizeof ( char ) * SizeX * SizeY * 3 );
	if ( !imgdes )
	{
		fprintf ( stderr, "Could not allocate enough memory for destination image\n" );
		exit ( -1 );

	}
	// read Data fromm raw file
	raw = fopen ( rawName, "r" );
	if ( raw == NULL )
	{
		fprintf ( stderr, "Could not open raw file\n" );
		exit ( -1 );
	}

	int result = fread ( imgscr,1,SizeX * SizeY * Bpp, raw );
	if ( result != SizeX*SizeY * Bpp )
	{
		fprintf ( stderr, "Could not read the right amount of byte from raw image\n" );
		fprintf ( stderr, "Expected %u , received %u\n", result,(unsigned int)(SizeX*SizeY * Bpp) );
		exit ( -1 );

	}
	// convert data if needed
	switch ( imgtype )
	{
		case RGB24: RGB24toRGB24 ( imgscr, imgdes,  SizeX,SizeY );
			break;

		case RGB32 : RGB32toRGB24 ( imgscr, imgdes, SizeX, SizeY );
			break;

		case RGB16 : RGB16toRGB24 ( imgscr, imgdes, SizeX, SizeY );
			break;

		case RGB15 : RGB15toRGB24 ( imgscr, imgdes, SizeX, SizeY );
			break;

		case Y8 : Y8toRGB24 ( imgscr, imgdes, SizeX, SizeY );
			break;

		case YUV422 : YUV422toRGB24 ( imgscr, imgdes, SizeX, SizeY );
			break;

		case YUV422PL : YUV422PLtoRGB24 ( imgscr, imgdes, SizeX, SizeY );
			break;

		case YUV420PL : YUV420PLtoRGB24 ( imgscr, imgdes, SizeX, SizeY );
			break;

	}



	// open the destination TIFF file
	if ( ( tif = TIFFOpen ( tifName, "w" ) ) == NULL )
	{
		printf ( "Could not open %s for writing\n",argv[4] );
		exit ( -1 );
	}
	// define Tiff tags
	TIFFSetField ( tif, TIFFTAG_IMAGEWIDTH, SizeX );
	TIFFSetField ( tif, TIFFTAG_IMAGELENGTH, SizeY );
	TIFFSetField ( tif, TIFFTAG_COMPRESSION, COMPRESSION_NONE );
	TIFFSetField ( tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG );
	TIFFSetField ( tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB );
	TIFFSetField ( tif, TIFFTAG_BITSPERSAMPLE, 8 );
	TIFFSetField ( tif, TIFFTAG_SAMPLESPERPIXEL, 3 );

	// put the data into the tiff
	if ( TIFFWriteEncodedStrip ( tif, 0, imgdes, SizeX * SizeY * 3 ) == 0 )
	{
		fprintf ( stderr, "Could not write image\n" );
		exit ( -2 );
	}

	// close the tiff
	TIFFClose ( tif );
	// close Raw file
	fclose ( raw );
        printf("\n Convertion done \n");
	return EXIT_SUCCESS;
}
